import dataclasses
from typing import Callable, List, Optional

from graphql import DirectiveLocation

from .utils.str_converters import to_camel_case


@dataclasses.dataclass
class DirectiveDefinition:
    name: str
    resolver: Callable
    locations: List[DirectiveLocation]
    description: Optional[str] = None


def directive(*, locations: List[DirectiveLocation], description=None, name=None):
    def _wrap(f):
        directive_name = name or to_camel_case(f.__name__)

        f.directive_definition = DirectiveDefinition(
            name=directive_name,
            locations=locations,
            description=description,
            resolver=f,
        )

        return f

    return _wrap



from typing import NoReturn, Optional, Tuple, Type


class StrawberryUnion:
    def __init__(
        self, name: str, types: Tuple[Type, ...], description: Optional[str] = None
    ):
        self.name = name
        self._types = types
        self.description = description

    @property
    def types(self) -> Tuple[Type, ...]:
        from .types.type_resolver import _resolve_generic_type

        types = tuple(
            _resolve_generic_type(t, self.name)
            for t in self._types
            if t is not None.__class__
        )

        return types

    def __call__(self, *_args, **_kwargs) -> NoReturn:
        """Do not use.

        Used to bypass
        https://github.com/python/cpython/blob/5efb1a77e75648012f8b52960c8637fc296a5c6d/Lib/typing.py#L148-L149
        """
        raise ValueError("Cannot use union type directly")


def union(
    name: str, types: Tuple[Type, ...], *, description: str = None
) -> StrawberryUnion:
    """Creates a new named Union type.

    Example usages:

    >>> strawberry.union("Some Thing", (int, str))

    >>> @strawberry.type
    ... class A: ...
    >>> @strawberry.type
    ... class B: ...
    >>> strawberry.union("Name", (A, Optional[B]))
    """

    union_definition = StrawberryUnion(name=name, types=types, description=description)

    return union_definition



import dataclasses
import inspect
from typing import Callable, List, Optional, Type

from .arguments import get_arguments_from_resolver
from .permission import BasePermission
from .types.types import FederationFieldParams, FieldDefinition
from .utils.str_converters import to_camel_case


class StrawberryField(dataclasses.Field):
    _field_definition: FieldDefinition

    def __init__(self, field_definition: FieldDefinition):
        self._field_definition = field_definition

        super().__init__(  # type: ignore
            default=dataclasses.MISSING,
            default_factory=dataclasses.MISSING,
            init=field_definition.base_resolver is None,
            repr=True,
            hash=None,
            compare=True,
            metadata=None,
        )

    def __call__(self, resolver: Callable) -> Callable:
        """Migrate the field definition to the resolver"""

        field_definition = self._field_definition
        # note that field_definition.name is finalized in type_resolver._get_fields

        field_definition.origin_name = resolver.__name__
        field_definition.origin = resolver
        field_definition.base_resolver = resolver
        field_definition.arguments = get_arguments_from_resolver(resolver)
        field_definition.type = resolver.__annotations__.get("return", None)

        if not inspect.ismethod(resolver):
            # resolver is a normal function
            resolver._field_definition = field_definition  # type: ignore
        else:
            # resolver is a bound method and immutable (most likely a
            # classmethod or an instance method). We need to monkeypatch its
            # underlying .__func__ function
            # https://stackoverflow.com/a/7891681/8134178
            resolver.__func__._field_definition = field_definition  # type:ignore

        return resolver

    def __setattr__(self, name, value):
        if name == "type":
            self._field_definition.type = value

        if value and name == "name":
            if not self._field_definition.origin_name:
                self._field_definition.origin_name = value

            if not self._field_definition.name:
                self._field_definition.name = to_camel_case(value)

        return super().__setattr__(name, value)


def field(
    resolver: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    federation: Optional[FederationFieldParams] = None
):
    """Annotates a method or property as a GraphQL field.

    This is normally used inside a type declaration:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = strawberry.field(description="ABC")

    >>>     @strawberry.field(description="ABC")
    >>>     def field_with_resolver(self, info) -> str:
    >>>         return "abc"

    it can be used both as decorator and as a normal function.
    """

    field_definition = FieldDefinition(
        origin_name=None,  # modified by resolver in __call__
        name=name,  # modified by resolver in __call__
        type=None,  # type: ignore
        origin=resolver,  # type: ignore
        description=description,
        base_resolver=resolver,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        arguments=[],  # modified by resolver in __call__
        federation=federation or FederationFieldParams(),
    )

    field_ = StrawberryField(field_definition)

    if resolver:
        return field_(resolver)
    return field_



from typing import Any, Dict, List, Optional

from graphql.error import format_error as format_graphql_error
from strawberry.types import ExecutionResult
from typing_extensions import TypedDict


class GraphQLHTTPResponse(TypedDict, total=False):
    data: Optional[Dict[str, Any]]
    errors: Optional[List[Any]]


def process_result(result: ExecutionResult) -> GraphQLHTTPResponse:
    data: GraphQLHTTPResponse = {"data": result.data}

    if result.errors:
        data["errors"] = [format_graphql_error(err) for err in result.errors]

    return data



from asyncio import ensure_future
from inspect import isawaitable
from typing import Any, Awaitable, Dict, List, Sequence, Type, cast

from graphql import (
    ExecutionResult as GraphQLExecutionResult,
    GraphQLError,
    GraphQLSchema,
    execute as original_execute,
    parse,
)
from graphql.type import validate_schema
from graphql.validation import validate
from strawberry.extensions import Extension
from strawberry.extensions.runner import ExtensionsRunner
from strawberry.types import ExecutionContext, ExecutionResult


async def execute(
    schema: GraphQLSchema,
    query: str,
    extensions: Sequence[Type[Extension]],
    root_value: Any = None,
    context_value: Any = None,
    variable_values: Dict[str, Any] = None,
    additional_middlewares: List[Any] = None,
    operation_name: str = None,
) -> ExecutionResult:
    execution_context = ExecutionContext(
        query=query,
        context=context_value,
        variables=variable_values,
        operation_name=operation_name,
    )

    extensions_runner = ExtensionsRunner(
        execution_context=execution_context,
        extensions=[extension() for extension in extensions],
    )

    additional_middlewares = additional_middlewares or []

    with extensions_runner.request():
        schema_validation_errors = validate_schema(schema)

        if schema_validation_errors:
            return ExecutionResult(
                data=None,
                errors=schema_validation_errors,
                extensions=extensions_runner.get_extensions_results(),
            )

        try:
            with extensions_runner.parsing():
                document = parse(query)
        except GraphQLError as error:
            return ExecutionResult(
                data=None,
                errors=[error],
                extensions=extensions_runner.get_extensions_results(),
            )

        except Exception as error:  # pragma: no cover
            error = GraphQLError(str(error), original_error=error)

            return ExecutionResult(
                data=None,
                errors=[error],
                extensions=extensions_runner.get_extensions_results(),
            )

        with extensions_runner.validation():
            validation_errors = validate(schema, document)

        if validation_errors:
            return ExecutionResult(data=None, errors=validation_errors)

        result = original_execute(
            schema,
            document,
            root_value=root_value,
            middleware=extensions_runner.as_middleware_manager(*additional_middlewares),
            variable_values=variable_values,
            operation_name=operation_name,
            context_value=context_value,
        )

        if isawaitable(result):
            result = await cast(Awaitable[GraphQLExecutionResult], result)

    result = cast(GraphQLExecutionResult, result)

    return ExecutionResult(
        data=result.data,
        errors=result.errors,
        extensions=extensions_runner.get_extensions_results(),
    )


def execute_sync(
    schema: GraphQLSchema,
    query: str,
    extensions: Sequence[Type[Extension]],
    root_value: Any = None,
    context_value: Any = None,
    variable_values: Dict[str, Any] = None,
    additional_middlewares: List[Any] = None,
    operation_name: str = None,
) -> ExecutionResult:
    execution_context = ExecutionContext(
        query=query,
        context=context_value,
        variables=variable_values,
        operation_name=operation_name,
    )
    extensions_runner = ExtensionsRunner(
        execution_context=execution_context,
        extensions=[extension() for extension in extensions],
    )

    additional_middlewares = additional_middlewares or []

    with extensions_runner.request():
        schema_validation_errors = validate_schema(schema)

        if schema_validation_errors:
            return ExecutionResult(
                data=None,
                errors=schema_validation_errors,
                extensions=extensions_runner.get_extensions_results(),
            )

        try:
            with extensions_runner.parsing():
                document = parse(query)
        except GraphQLError as error:
            return ExecutionResult(
                data=None,
                errors=[error],
                extensions=extensions_runner.get_extensions_results(),
            )

        except Exception as error:  # pragma: no cover
            error = GraphQLError(str(error), original_error=error)

            return ExecutionResult(
                data=None,
                errors=[error],
                extensions=extensions_runner.get_extensions_results(),
            )

        with extensions_runner.validation():
            validation_errors = validate(schema, document)

        if validation_errors:
            return ExecutionResult(data=None, errors=validation_errors)

        result = original_execute(
            schema,
            document,
            root_value=root_value,
            middleware=extensions_runner.as_middleware_manager(*additional_middlewares),
            variable_values=variable_values,
            operation_name=operation_name,
            context_value=context_value,
        )

        if isawaitable(result):
            ensure_future(cast(Awaitable[GraphQLExecutionResult], result)).cancel()
            raise RuntimeError("GraphQL execution failed to complete synchronously.")

    result = cast(GraphQLExecutionResult, result)

    return ExecutionResult(
        data=result.data,
        errors=result.errors,
        extensions=extensions_runner.get_extensions_results(),
    )



from typing import Callable, Optional, Type, Union, cast

from graphql import GraphQLList, GraphQLNonNull, GraphQLType
from strawberry.field import FieldDefinition
from strawberry.scalars import is_scalar
from strawberry.types.types import ArgumentDefinition
from strawberry.union import StrawberryUnion

from .enum import get_enum_type
from .scalar import get_scalar_type
from .types import TypeMap
from .union import get_union_type


def get_type_for_annotation(annotation: Type, type_map: TypeMap) -> GraphQLType:
    graphql_type: Optional[GraphQLType] = None

    # this adds support for type annotations that use
    # strings, without having to use get_type_hints

    if type(annotation) == str and annotation in type_map:
        return type_map[annotation].implementation  # type: ignore

    if is_scalar(annotation):
        graphql_type = get_scalar_type(annotation, type_map)

    elif hasattr(annotation, "_enum_definition"):
        graphql_type = get_enum_type(annotation._enum_definition, type_map)
    elif hasattr(annotation, "_type_definition"):
        from .object_type import get_object_type

        graphql_type = get_object_type(annotation, type_map)

    if not graphql_type:
        raise ValueError(f"Unable to get GraphQL type for {annotation}")

    return graphql_type


def get_graphql_type(
    field: Union[FieldDefinition, ArgumentDefinition], type_map: TypeMap,
) -> GraphQLType:
    # by default fields in GraphQL-Core are optional, but for us we only want
    # to mark optional fields when they are inside a Optional type hint
    wrap: Optional[Callable] = GraphQLNonNull

    type: GraphQLType
    field_type = cast(Type, field.type)

    if field.is_optional:
        wrap = None

    if field.is_list:
        child = cast(FieldDefinition, field.child)
        type = GraphQLList(get_graphql_type(child, type_map))

    elif field.is_union:
        union_definition = cast(StrawberryUnion, field_type)
        type = get_union_type(union_definition, type_map)
    else:
        type = get_type_for_annotation(field_type, type_map)

    if wrap:
        return wrap(type)

    return type



from typing import Callable, Optional

from mypy.nodes import (
    GDEF,
    Expression,
    IndexExpr,
    NameExpr,
    SymbolTableNode,
    TupleExpr,
    TypeAlias,
)
from mypy.plugin import (
    AnalyzeTypeContext,
    ClassDefContext,
    DynamicClassDefContext,
    Plugin,
    SemanticAnalyzerPluginInterface,
)
from mypy.plugins import dataclasses
from mypy.types import Type, UnionType


def lazy_type_analyze_callback(ctx: AnalyzeTypeContext) -> Type:
    type_name = ctx.type.args[0]
    type_ = ctx.api.analyze_type(type_name)

    return type_


def private_type_analyze_callback(ctx: AnalyzeTypeContext) -> Type:
    type_name = ctx.type.args[0]
    type_ = ctx.api.analyze_type(type_name)

    return type_


def _get_type_for_expr(expr: Expression, api: SemanticAnalyzerPluginInterface):
    if isinstance(expr, NameExpr):
        return api.named_type(expr.name)

    if isinstance(expr, IndexExpr):
        type_ = _get_type_for_expr(expr.base, api)
        type_.args = [_get_type_for_expr(expr.index, api)]

        return type_

    raise ValueError(f"Unsupported expression f{type(expr)}")


def union_hook(ctx: DynamicClassDefContext) -> None:
    types = ctx.call.args[1]

    if isinstance(types, TupleExpr):
        type_ = UnionType(tuple(_get_type_for_expr(x, ctx.api) for x in types.items))

        type_alias = TypeAlias(
            type_,
            fullname=ctx.api.qualified_name(ctx.name),
            line=ctx.call.line,
            column=ctx.call.column,
        )

        ctx.api.add_symbol_table_node(
            ctx.name, SymbolTableNode(GDEF, type_alias, plugin_generated=False)
        )


class StrawberryPlugin(Plugin):
    def get_dynamic_class_hook(
        self, fullname: str
    ) -> Optional[Callable[[DynamicClassDefContext], None]]:
        # TODO: investigate why we need this instead of `strawberry.union.union` on CI
        # we have the same issue in the other hooks
        if "strawberry.union" in fullname:
            return union_hook

        return None

    def get_type_analyze_hook(self, fullname: str):
        if fullname == "strawberry.lazy_type.LazyType":
            return lazy_type_analyze_callback

        if any(
            name in fullname
            for name in {"strawberry.private.Private", "strawberry.Private"}
        ):
            return private_type_analyze_callback

        return None

    def get_class_decorator_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]:
        if any(
            strawberry_decorator in fullname
            for strawberry_decorator in {
                "strawberry.type",
                "strawberry.federation.type",
                "strawberry.input",
                "strawberry.interface",
            }
        ):
            return dataclasses.dataclass_class_maker_callback
        return None


def plugin(version: str):
    return StrawberryPlugin



import json
import os
from typing import Any, Dict, Optional

from django.http import Http404, HttpRequest, HttpResponseNotAllowed, JsonResponse
from django.http.response import HttpResponseBadRequest
from django.template import RequestContext, Template
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

import strawberry
from strawberry.file_uploads.data import replace_placeholders_with_files
from strawberry.http import GraphQLHTTPResponse, process_result
from strawberry.types import ExecutionResult

from ..schema import BaseSchema


class GraphQLView(View):
    graphiql = True
    schema: Optional[BaseSchema] = None

    def __init__(self, schema: BaseSchema, graphiql=True):
        self.schema = schema
        self.graphiql = graphiql

    def get_root_value(self, request: HttpRequest) -> Any:
        return None

    def get_context(self, request: HttpRequest) -> Any:
        return {"request": request}

    def parse_body(self, request) -> Dict[str, Any]:
        if request.content_type == "multipart/form-data":
            data = json.loads(request.POST.get("operations", "{}"))
            files_map = json.loads(request.POST.get("map", "{}"))

            data = replace_placeholders_with_files(data, files_map, request.FILES)

            return data

        return json.loads(request.body)

    def process_result(
        self, request: HttpRequest, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        return process_result(result)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        if request.method.lower() not in ("get", "post"):
            return HttpResponseNotAllowed(
                ["GET", "POST"], "GraphQL only supports GET and POST requests."
            )

        if "text/html" in request.META.get("HTTP_ACCEPT", ""):
            if not self.graphiql:
                raise Http404("GraphiQL has been disabled")

            return self._render_graphiql(request)

        data = self.parse_body(request)

        try:
            query = data["query"]
            variables = data.get("variables")
            operation_name = data.get("operationName")
        except KeyError:
            return HttpResponseBadRequest("No GraphQL query found in the request")

        context = self.get_context(request)

        result = self.schema.execute_sync(
            query,
            root_value=self.get_root_value(request),
            variable_values=variables,
            context_value=context,
            operation_name=operation_name,
        )

        response_data = self.process_result(request=request, result=result)

        return JsonResponse(response_data)

    def _render_graphiql(self, request, context=None):
        try:
            template = Template(render_to_string("graphql/graphiql.html"))
        except TemplateDoesNotExist:
            template = Template(
                open(
                    os.path.join(
                        os.path.dirname(os.path.abspath(strawberry.__file__)),
                        "static/graphiql.html",
                    ),
                    "r",
                ).read()
            )

        context = context or {}
        context.update({"SUBSCRIPTION_ENABLED": "false"})

        response = TemplateResponse(request=request, template=None, context=context)
        response.content = template.render(RequestContext(request, context))

        return response



from typing import Type, cast

from graphql import GraphQLInputObjectType, GraphQLObjectType
from graphql.type.definition import GraphQLInterfaceType
from strawberry.type import TypeDefinition

from .fields import get_field
from .types import ConcreteType, GraphQLType, TypeMap


def _get_object_type_for_type_definition(
    type_definition: TypeDefinition, type_map: TypeMap
) -> GraphQLType:

    if type_definition.name not in type_map:
        TypeClass: Type = GraphQLObjectType

        kwargs = {}

        if type_definition.is_input:
            TypeClass = GraphQLInputObjectType
        elif type_definition.is_interface:
            TypeClass = GraphQLInterfaceType

        if type_definition.interfaces:
            kwargs["interfaces"] = [
                _get_object_type_for_type_definition(interface, type_map)
                for interface in type_definition.interfaces
            ]
            # this tells GraphQL core what the returned object's actual type is
            kwargs["is_type_of"] = lambda obj, _: isinstance(  # type: ignore
                obj, type_definition.origin
            )

        assert not type_definition.is_generic

        object_type = TypeClass(
            name=type_definition.name,
            fields=lambda: {
                field.name: get_field(field, type_definition.is_input, type_map)
                for field in type_definition.fields
            },
            description=type_definition.description,
            **kwargs,
        )

        type_map[type_definition.name] = ConcreteType(
            definition=type_definition, implementation=object_type
        )

    return type_map[type_definition.name].implementation


def get_object_type(origin: Type, type_map: TypeMap) -> GraphQLObjectType:
    """Returns a root type (Query, Mutation, Subscription) from a decorated type"""

    if not hasattr(origin, "_type_definition"):
        raise ValueError(f"Wrong type passed to get object type {origin}")

    type_definition: TypeDefinition = origin._type_definition

    return cast(
        GraphQLObjectType,
        _get_object_type_for_type_definition(type_definition, type_map),
    )



from functools import partial

from .field import field


# Mutations and subscriptions are field, we might want to separate things in the long run
# for example to provide better errors
mutation = field
subscription = partial(field, is_subscription=True)



import datetime
import decimal
from operator import methodcaller

from strawberry.custom_scalar import scalar


isoformat = methodcaller("isoformat")


Date = scalar(
    datetime.date,
    name="Date",
    description="Date (isoformat)",
    serialize=isoformat,
    parse_value=datetime.date.fromisoformat,
)
DateTime = scalar(
    datetime.datetime,
    name="DateTime",
    description="Date with time (isoformat)",
    serialize=isoformat,
    parse_value=datetime.datetime.fromisoformat,
)
Time = scalar(
    datetime.time,
    name="Time",
    description="Time (isoformat)",
    serialize=isoformat,
    parse_value=datetime.time.fromisoformat,
)

Decimal = scalar(
    decimal.Decimal,
    name="Decimal",
    description="Decimal (fixed-point)",
    serialize=str,
    parse_value=decimal.Decimal,
)



import builtins
import dataclasses
from typing import Dict, Iterable, Tuple, Type, cast

from strawberry.union import StrawberryUnion, union
from strawberry.utils.str_converters import capitalize_first
from strawberry.utils.typing import is_type_var

from .types import FederationFieldParams, FieldDefinition, TypeDefinition


def get_name_from_types(types: Iterable[Type]):
    return "".join([capitalize_first(type.__name__) for type in types])


def copy_type_with(
    base: Type, *types: Type, params_to_type: Dict[Type, Type] = None
) -> Type:
    if params_to_type is None:
        params_to_type = {}

    if isinstance(base, StrawberryUnion):
        types = cast(
            Tuple[Type, ...],
            tuple(copy_type_with(t, params_to_type=params_to_type) for t in base.types),
        )

        return union(
            name=get_name_from_types(types), types=types, description=base.description,
        )

    if hasattr(base, "_type_definition"):
        definition = cast(TypeDefinition, base._type_definition)

        if definition.type_params:
            fields = []

            type_params = definition.type_params.values()
            params_to_type.update(dict(zip(type_params, types)))

            name = get_name_from_types(params_to_type.values()) + definition.name

            for field in definition.fields:
                kwargs = dataclasses.asdict(field)

                if field.is_list:
                    child = cast(FieldDefinition, field.child)
                    child_type = cast(Type, child.type)

                    # TODO: nested list

                    kwargs["child"] = FieldDefinition(
                        name=child.name,
                        origin=child.origin,
                        origin_name=child.origin_name,
                        is_optional=child.is_optional,
                        type=copy_type_with(child_type, params_to_type=params_to_type),
                    )

                else:
                    field_type = cast(Type, field.type)

                    kwargs["type"] = copy_type_with(
                        field_type, params_to_type=params_to_type
                    )

                federation_args = kwargs.pop("federation")
                kwargs["federation"] = FederationFieldParams(**federation_args)

                fields.append(FieldDefinition(**kwargs))

            type_definition = TypeDefinition(
                name=name,
                is_input=definition.is_input,
                origin=definition.origin,
                is_interface=definition.is_interface,
                is_generic=False,
                federation=definition.federation,
                interfaces=definition.interfaces,
                description=definition.description,
                _fields=fields,
            )
            type_definition._type_params = {}

            copied_type = builtins.type(
                name, (), {"_type_definition": type_definition},
            )

            if not hasattr(base, "_copies"):
                base._copies = {}

            base._copies[types] = copied_type

            return copied_type

    if is_type_var(base):
        return params_to_type[base]

    return base



import enum
import inspect
from typing import Any, Callable, Dict, List, Mapping, Type, cast

from .exceptions import MissingArgumentsAnnotationsError, UnsupportedTypeError
from .scalars import is_scalar
from .types.type_resolver import resolve_type
from .types.types import ArgumentDefinition, undefined
from .utils.str_converters import to_camel_case


def get_arguments_from_annotations(
    annotations: Any, parameters: Mapping[str, inspect.Parameter], origin: Any
) -> List[ArgumentDefinition]:
    arguments = []

    for name, annotation in annotations.items():
        default_value = parameters[name].default
        default_value = (
            undefined
            if default_value in (inspect._empty, None)  # type: ignore
            else default_value
        )

        argument_definition = ArgumentDefinition(
            origin_name=name,
            name=to_camel_case(name),
            origin=origin,
            type=annotation,
            default_value=default_value,
        )

        arguments.append(argument_definition)

        resolve_type(argument_definition)

    return arguments


def get_arguments_from_resolver(resolver: Callable) -> List[ArgumentDefinition]:
    annotations = resolver.__annotations__
    parameters = inspect.signature(resolver).parameters
    function_arguments = set(parameters) - {"root", "self", "info"}

    annotations = {
        name: annotation
        for name, annotation in annotations.items()
        if name not in ["root", "info", "return", "self"]
    }

    annotated_function_arguments = set(annotations.keys())
    arguments_missing_annotations = function_arguments - annotated_function_arguments

    if len(arguments_missing_annotations) > 0:
        raise MissingArgumentsAnnotationsError(
            resolver.__name__, arguments_missing_annotations
        )

    return get_arguments_from_annotations(annotations, parameters, origin=resolver)


class _Unset:
    def __str__(self):
        return ""

    def __bool__(self):
        return False


UNSET: Any = _Unset()


def is_unset(value: Any) -> bool:
    return type(value) is _Unset


def convert_argument(value: Any, argument_definition: ArgumentDefinition) -> Any:
    if value is None:
        return None

    if is_unset(value):
        return value

    if argument_definition.is_list:
        child_definition = cast(ArgumentDefinition, argument_definition.child)

        return [convert_argument(x, child_definition) for x in value]

    argument_type = cast(Type, argument_definition.type)

    if is_scalar(argument_type):
        return value

    # Convert Enum fields to instances using the value. This is safe
    # because graphql-core has already validated the input.
    if isinstance(argument_type, enum.EnumMeta):
        return argument_type(value)  # type: ignore

    if hasattr(argument_type, "_type_definition"):
        assert argument_type._type_definition.is_input

        kwargs = {}

        for field in argument_type._type_definition.fields:
            if field.name in value:
                kwargs[field.origin_name] = convert_argument(value[field.name], field)

        return argument_type(**kwargs)

    raise UnsupportedTypeError(argument_type)


def convert_arguments(
    value: Dict[str, Any], arguments: List[ArgumentDefinition],
) -> Dict[str, Any]:
    """Converts a nested dictionary to a dictionary of actual types.

    It deals with conversion of input types to proper dataclasses and
    also uses a sentinel value for unset values."""

    if not arguments:
        return {}

    kwargs = {}

    for argument in arguments:
        if argument.name in value:
            origin_name = cast(str, argument.origin_name)
            current_value = value[argument.name]

            kwargs[origin_name] = convert_argument(current_value, argument)

    return kwargs



import datetime
import decimal
from typing import Dict, Type, cast
from uuid import UUID

from graphql import (
    GraphQLBoolean,
    GraphQLFloat,
    GraphQLID,
    GraphQLInt,
    GraphQLScalarType,
    GraphQLString,
)
from strawberry.custom_scalar import SCALAR_REGISTRY, ScalarDefinition
from strawberry.file_uploads.scalars import Upload
from strawberry.scalars import ID

from .base_scalars import Date, DateTime, Decimal, Time
from .types import ConcreteType, TypeMap


def _make_scalar_type(definition: ScalarDefinition) -> GraphQLScalarType:
    return GraphQLScalarType(
        name=definition.name,
        description=definition.description,
        serialize=definition.serialize,
        parse_value=definition.parse_value,
        parse_literal=definition.parse_literal,
    )


DEFAULT_SCALAR_REGISTRY: Dict[Type, GraphQLScalarType] = {
    str: GraphQLString,
    int: GraphQLInt,
    float: GraphQLFloat,
    bool: GraphQLBoolean,
    ID: GraphQLID,
    UUID: GraphQLID,
    Upload: _make_scalar_type(Upload._scalar_definition),
    datetime.date: _make_scalar_type(Date._scalar_definition),
    datetime.datetime: _make_scalar_type(DateTime._scalar_definition),
    datetime.time: _make_scalar_type(Time._scalar_definition),
    decimal.Decimal: _make_scalar_type(Decimal._scalar_definition),
}


def get_scalar_type(annotation: Type, type_map: TypeMap) -> GraphQLScalarType:
    if annotation in DEFAULT_SCALAR_REGISTRY:
        return DEFAULT_SCALAR_REGISTRY[annotation]

    if annotation in SCALAR_REGISTRY:
        scalar_definition = SCALAR_REGISTRY[annotation]
    else:
        scalar_definition = annotation._scalar_definition

    if scalar_definition.name not in type_map:
        type_map[scalar_definition.name] = ConcreteType(
            definition=scalar_definition,
            implementation=_make_scalar_type(scalar_definition),
        )

    return cast(GraphQLScalarType, type_map[scalar_definition.name].implementation)
